/*****************************************************************************
 * Orangutan LV-168 Demo program 3:                                          *
 * LCD, user trimpot, Analog-to-Digital Conversion (ADC), and motor control. *
 *****************************************************************************/

// This program uses the mega168's analog-to-digital converter to monitor the
// position of the user trimpot.  It displays the trimpot's position (a number
// from 0 to 1023) on the LCD and drives its two motor ports based on this
// position.  Motor speed will be zero for a trimpot position of 511, maximum
// "forward" for a trimpot position of zero, and maximum "reverse" for a 
// trimpot position of 1023.


#include <avr/io.h>
#include "device.h"		// pinout definitions for the Orangutan LV-168
#include <util/delay.h>	// F_CPU is defined in "device.h" above

#include "lcd.h"		// provides routines for using the LCD


// ***** MOTOR PINS *****
#define M1B					PD5
#define M1A					PD6
#define M2B					PD3
#define M2A					PB3		// *** NOTE: M2A is on port B, not port D


// ***** MOTOR CONTROL MACROS *****
// drive M1A high and pwm M1B, alternating between M1 full forward and brake
#define M1_FORWARD(pwm)		OCR0A = 0; OCR0B = pwm

// drive M1B high and pwm M1A, alternating between M1 full reverse and brake
#define M1_REVERSE(pwm)		OCR0B = 0; OCR0A = pwm

// drive M2A high and pwm M2B, alternating between M2 full forward and brake
#define M2_FORWARD(pwm)		OCR2A = 0; OCR2B = pwm

// drive M2B high and pwm M2A, alternating between M2 full reverse and brake
#define M2_REVERSE(pwm)		OCR2B = 0; OCR2A = pwm


// ***** MOTOR PWM INITIALIZATION *****
void motors_init()
{
	// set the output state of the four motor pins to high
	//  (this will enable their internal pullups if they are inputs
	//  or drive them high if they are outputs)
	PORTD |= (1 << M1A) | (1 << M1B) | (1 << M2B);
	PORTB |= (1 << M2A);

	// and make them outputs (in this state their output values will be
	//  overriden by PWM outputs once timers 0 and 2 are correctly configured)
	DDRD |= (1 << M1A) | (1 << M1B) | (1 << M2B);
	DDRB |= (1 << M2A);

	// Configure timer 0 and timer 2 to generate fast PWMs at a frequency of
	// 9.77 kHz.  Registers OCR0A, OCR0B, OCR2A, and OCR2B are used to control
	// the duty cycles of the PWMs that drive the motors.  Timer 0 PWMs are
	// used to control motor 1 and timer 2 PWMs are used to control motor 2.
	// The general idea is to hold one side of the H-bridge high while
	// PWMing the other side, which causes the motor to alternate between
	// drive (one side high, the other side low) and brake (both sides high).
	// See the motor control macros at the top of this file for more details.
	OCR0A = OCR0B = OCR2A = OCR2B = 0;	// set duty cycles initially for brake
	TCCR0A = TCCR2A = 0xF3;			// set OCxx on compare match, clear at TOP
									//  with WGM02=0, select fast PWM, TOP=0xFF
	TCCR0B = TCCR2B = 0x02;			// WGM02=0, timer clock = IO clk / 8
									//  => PMW freq = 20MHz/8/256 = 9.77 kHz
}


// delay for time_ms milliseconds by looping
void delay_ms( unsigned int time_ms )
{
	unsigned int i;

	for ( i = 0; i < time_ms; i++ )
		_delay_ms( 1 );		// _delay_ms() comes from <util/delay.h>
							//  and can only delay for a max of around 13 ms
							//  when the IO clock is 20 MHz.
}


// Controlling an LCD can be somewhat complicated, but you can get around that
// by making use of LCD code that is already written for you.  In this demo we
// are using functions defined in "lcd.c".  Simply put "lcd.c", "lcd.h", and
// "device.h" in the same directory as your AVR Studio project files.  Add
// "lcd.c" to your project and include "lcd.h" in any project file that needs
// to access lcd functions.  "lcd.c" was written by Tom Benedict as part of
// OrangutanLib, an open source set of C libraries for the Orangutan family of
// robot controllers.  We have added a few more functions to the version of
// "lcd.c" included as part of this demo and have modified "device.h" to
// contain the correct pinout information for the Orangutan LV-168.

// You may notice that the green LED faintly glows while this program is
// running, even though we don't use the LEDs at all.  This is because the
// green LED IO line is also used to send data to the LCD.  When data is sent
// to the LCD, the LCD temporarily takes over use of the line before restoring
// it to it's previous state.  It is possible to use both the green LED and
// LCD simultaneously, though the LCD will prevent the green LED from ever
// going fully off or fully on.

// NOTE: if this is your first time using the LCD and you don't see any text
//  appear, you might need to adjust the LCD backlight using the LCD trimpot
//  in the upper-right corner of the board.  Typically you will need to rotate
//  this trimpot clockwise as far as it will go.


int main()
{
	lcd_init();	// this function must be called before any other LCD command

	lcd_string("OrgLV168");	// display the string on the LCD
	lcd_gotoxy(0, 1);		// goto the start of LCD line 2
	lcd_string(" Demo 3");

	delay_ms(1500);			// loop-delay for 1.5 seconds
	lcd_clear();


	motors_init();


	// ***** ADC INITIALIZATION *****

	// The user trimpot can be optionally jumpered to ADC7 (a dedicated
	// analog input on the mega168 referred to as ADC channel 7) using a blue
	// shorting block; the Orangutan LV-168 ships with this shorting
	// block in place.  This program assumes that the trimpot is connected
	// to ADC7 via this shorting block.

	ADCSRA = 0x87;		// bit 7 set: ADC enabled
						// bit 6 clear: don't start conversion
						// bit 5 clear: disable autotrigger
						// bit 4: ADC interrupt flag
						// bit 3 clear: disable ADC interrupt
						// bits 0-2 set: ADC clock prescaler is 128

	ADMUX = 0x07;		// bit 7 and 6 clear: voltage ref is Vref pin
						// bit 5 clear: right-adjust result (10-bit ADC)
						// bit 4 not implemented
						// bits 0-3: ADC channel (channel 7)


	while (1)			// loop forever
	{
		long sum = 0;
		unsigned int avg, i;

		// Here we accumulate 512 conversion results to get an average ADC.
		// According to the mega168 datasheet, it takes approximately 13
		// ADC clock cycles to perform a conversion.  We have configured
		// the ADC run at IO clock / 128 = 20 MHz / 128 = 156 kHz, which
		// means it has a conversion rate of around 10 kHz.  As a result,
		// it should take around 50 ms to accumulate 500 ADC samples.
		for (i = 0; i < 512; i++)
		{
			ADCSRA |= ( 1 << ADSC );			// start conversion
			while ( ADCSRA & ( 1 << ADSC ))		// wait while converting
				;
			sum += ADC;							// add in conversion result
		}
		avg = sum >> 9;		// a.k.a. avg = sum / 2^9 = avg / 512
							//  typically, bitshifting is faster than division

		lcd_gotoxy(0, 0);				// go to the start of LCD line 1
		lcd_string("Trimpot");
		lcd_gotoxy(0, 1);				// go to the start of LCD line 2
		lcd_int(avg);					// display average ADC as an integer
		lcd_string("   ");				// overwrite any LCD character remnants


		// set motors based on trimpot position (middle value = motor speed 0)
		if (avg <= 511)
		{
			M1_FORWARD((511 - avg) >> 1);	// M1 speed = (511 - avg) / 2
			M2_FORWARD((511 - avg) >> 1);	// M2 speed = (511 - avg) / 2
		}
		else
		{
			M1_REVERSE((avg - 512) >> 1);	// M1 speed = (avg - 512) / 2
			M2_REVERSE((avg - 512) >> 1);	// M2 speed = (avg - 512) / 2
		}
	}


	return 0;
}
